from __future__ import print_function
from collections import OrderedDict
import copy

from .. import pawstools

class PawsPlugin(object):
    """Base class for building PAWS Plugins."""

    def __init__(self,inputs):
        super(PawsPlugin,self).__init__()
        self.inputs = OrderedDict(copy.deepcopy(inputs))
        self.input_locator = OrderedDict.fromkeys(self.inputs.keys())
        self.input_doc = OrderedDict.fromkeys(self.inputs.keys()) 
        self.message_callback = self.tagged_print
        self.data_callback = None 
        self.running = False
        for name in self.inputs.keys(): 
            self.input_locator[name] = pawstools.InputLocator(pawstools.basic_type,self.inputs[name])

    def __getitem__(self,key):
        if key == 'inputs':
            return self.inputs
        else:
            raise KeyError('[{}] {} not in valid plugin keys: {}'
            .format(__name__,key,self.keys()))
    def keys(self):
        return ['inputs'] 

    def tagged_print(self,msg):
        print('[{}] {}'.format(type(self).__name__,msg))

    def get_plugin_content(self):
        return {}

    def description(self):
        """Describe the plugin.

        PawsPlugin.description() returns a string 
        documenting the functionality of the PawsPlugin,
        the current input settings, etc.
        Reimplement this in PawsPlugin subclasses.
        """
        return str(self.setup_dict()) 

    def start(self):
        """Start the plugin.

        Assuming a plugin's inputs have been set,
        PawsPlugin.start() should prepare the plugin for use, 
        e.g. by opening connections, reading files, etc. 
        Reimplement this in PawsPlugin subclasses as needed.
        """
        self.running = True 

    def stop(self):
        """Stop the plugin.

        PawsPlugin.stop() should provide a clean end for the plugin,
        for instance closing connections used by the plugin.
        Reimplement this in PawsPlugin subclasses as needed.
        """
        self.running = False 

    @classmethod
    def clone(cls):
        return cls()

    def build_clone(self):
        """Clone the Plugin."""
        new_pgn = self.clone()
        for inp_nm,il in self.input_locator.items():
            if il.tp == pawstools.basic_type:
                new_pgn.inputs[inp_nm] = copy.deepcopy(self.inputs[inp_nm]) 
            elif il.tp == pawstools.plugin_item:
                # plugins are expected to be threadsafe
                new_pgn.inputs[inp_nm] = self.inputs[inp_nm]
        #new_pgn.data_callback = self.data_callback
        #new_pgn.message_callback = self.message_callback
        #if self.running:
        #    new_pgn.start()
        return new_pgn

    def setup_dict(self):
        """Return a dict that states the plugin's module and inputs."""
        pgin_mod = self.__module__[self.__module__.find('plugins'):]
        pgin_mod = pgin_mod[pgin_mod.find('.')+1:]
        dct = OrderedDict() 
        dct['plugin_module'] = pgin_mod
        inp_dct = OrderedDict() 
        for nm,il in self.input_locator.items():
            inp_dct[nm] = {'tp':copy.copy(il.tp),'val':copy.copy(il.val)}
        dct['inputs'] = inp_dct 
        return dct





from collections import OrderedDict

import numpy as np

from ...Operation import Operation
from ... import optools

inputs=OrderedDict(
    batch_outputs=None,
    output_key=None)
outputs=OrderedDict(data_list=None)

class ListFromBatch(Operation):
    """
    Given a batch output (list of dicts)
    and an output name (dict key), 
    harvest a list of outputs from the batch.
    """

    def __init__(self):
        super(ListFromBatch,self).__init__(inputs,outputs)        
        self.input_doc['batch_outputs'] = 'list of dicts produced by a batch execution'
        self.input_doc['output_key'] = 'name of workflow output to be harvested'
        self.output_doc['data_list'] = 'list of the data fetched from batch_output'

    def run(self):
        b_out = self.inputs['batch_outputs']
        k = self.inputs['output_key']
        self.outputs['data_list'] = [d[k] for d in b_out]




from collections import OrderedDict
import copy

import os
import numpy as np

from ...Operation import Operation

inputs=OrderedDict(
    array=None,
    headers=None,
    dir_path=None,
    filename=None,
    filetag='')
outputs=OrderedDict(
    file_path=None,
    filename=None)

class WriteArrayCSV(Operation):
    """Write a 2d array to a csv file"""

    def __init__(self):
        super(WriteArrayCSV, self).__init__(inputs, outputs)
        self.input_doc['array'] = 'any 2d array'
        self.input_doc['headers'] = 'list of string headers (optional)- one header for each column of array'
        self.input_doc['dir_path'] = 'the path to the destination directory'
        self.input_doc['filename'] = 'the name of the file to be saved- no extension is expected'
        self.input_doc['filetag'] = 'tag appended to filename- no extension is expected'
        self.output_doc['file_path'] = 'the path to the finished csv file: dir_path+filename+filetag+.csv'
        self.output_doc['file_path'] = 'the name of the output csv: filename+filetag'
        self.input_datatype['headers'] = 'list'
        self.input_datatype['dir_path'] = 'str'
        self.input_datatype['filename'] = 'str'
        self.input_datatype['filetag'] = 'str'

    def run(self):
        a = self.inputs['array']
        h = self.inputs['headers']
        p = self.inputs['dir_path']
        fnm = self.inputs['filename']
        tag = self.inputs['filetag']
        csv_path = os.path.join(p,self.inputs['filename']+tag)
        if not os.path.splitext(csv_path)[1] == '.csv':
            csv_path += '.csv'
        self.outputs['file_path'] = csv_path
        self.outputs['filename'] = fnm+tag 

        if h is not None:
            h_str = ''
            for i in range(len(h)-1):
                h_str += h[i] + ', '
            h_str = h_str+h[-1]
            np.savetxt(csv_path, a, delimiter=',', newline=os.linesep, header=h_str)
        else:
            np.savetxt(csv_path, a, delimiter=',', newline=os.linesep)




import os
from collections import OrderedDict

import fabio 

from ...Operation import Operation

inputs=OrderedDict(file_path=None)
outputs=OrderedDict(
    image_data=None,
    header_dict=None,
    dir_path=None,
    filename=None)

class ReadImageAndHeader_SSRL15(Operation):
    """Read an image and header from SSRL beam line 1-5."""

    def __init__(self):
        super(ReadImageAndHeader_SSRL15, self).__init__(inputs, outputs)
        self.input_doc['file_path'] = 'path to a tif file '\
            'produced by beamline 1-5 at SSRL. '\
            'A .txt header file is expected '\
            'in the same directory as this .tif file.'
        self.output_doc['image_data'] = 'the image data as an ndarray'
        self.output_doc['header_dict'] = 'the header data as a python dictionary'
        self.output_doc['dir_path'] = 'the directory portion of the input file_path'
        self.output_doc['filename'] = 'filename with path and extension stripped'

    def run(self):
        p = self.inputs['file_path']
        dirpath,filename = os.path.split(p)
        filename_noext = os.path.splitext(filename)[0]
        path_noext = os.path.splitext(p)[0]
        hdr_file_path = path_noext + '.txt'
        self.outputs['dir_path'] = dirpath 
        self.outputs['filename'] = filename_noext 
        self.outputs['image_data'] = fabio.open(p).data
        d = OrderedDict()
        for l in open(hdr_file_path,'r').readlines():
            if not l.strip() == '' and not l.strip()[0] == '#':
                kvs = l.split(',')
                # special case for the string headers on line 1
                if not kvs[0].find('User') == -1:
                    u_str = kvs[0].split('User:')[1].strip()
                    t_str = kvs[1].split('time:')[1].strip()
                    d['User'] = u_str
                    d['time'] = t_str
                # and filter out the redundant temperature line
                elif not (len(kvs)==1 and kvs[0].strip()[-1]=='C'):
                    for kv in kvs:
                        kv_arr = kv.split('=')
                        d[kv_arr[0].strip()] = float(kv_arr[1].strip())
            self.outputs['header_dict'] = d




from __future__ import print_function
from collections import OrderedDict
from functools import partial
import copy
#from multiprocessing import Process,Pool

from ..operations.OpManager import OpManager
from ..plugins.PluginManager import PluginManager
from .Workflow import Workflow
from .. import pawstools

class WfManager(object):
    """
    Manager for paws Workflows. 
    Stores a list of Workflow objects, 
    performs operations on them.
    Keeps a reference to a PluginManager
    for access to PawsPlugins.
    """

    def __init__(self,op_manager=None,plugin_manager=None):
        """Initialize a workflow manager.

        Parameters
        ----------
        op_manager : OpManager (optional)
            an operations manager (see paws.core.operations.OpManager)-
            if not provided, a default OpManager will be created.
        plugin_manager : PluginManager (optional)
            a plugins manager (see paws.core.plugins.PluginManager)-
            if not provided, a default PluginManager will be created.
        """
        super(WfManager,self).__init__()
        if not op_manager:
            op_manager = OpManager()
        if not plugin_manager:
            plugin_manager = PluginManager()
        self.op_manager = op_manager 
        self.plugin_manager = plugin_manager 
        self.workflows = OrderedDict() 
        # dict of workflow clones for executing across threads:
        self.wf_clones = OrderedDict()
        # dict of bools to keep track of who is at work:
        self.wf_running = OrderedDict() 
        self.plugin_names = OrderedDict()
        self.message_callback = self.tagged_print
        self.pool=None

    def tagged_print(self,msg):
        print('[{}] {}'.format(type(self).__name__,msg))

    def set_pool(self,pool):
        self.pool = pool
        self.plugin_manager.pool = pool

    def add_workflow(self,wf_name):
        """Name and add a workflow.

        If `wf_name` is not unique (i.e. a workflow with that name already exists),
        this method will overwrite the existing workflow with a new one.

        Parameters
        ----------
        wf_name : str
            name to give to the new Workflow

        Returns
        -------
        wf : Workflow
            a reference to the new Workflow
        """
        wf = Workflow()
        if not wf.is_tag_valid(wf_name): 
            raise pawstools.WfNameError(wf.tag_error_message(wf_name))
        #wf.message_callback = self.message_callback
        self.workflows[wf_name] = wf
        self.wf_running[wf_name] = False
        return wf

    def n_workflows(self):
        """Return the current number of Workflows"""
        return len(self.workflows)

    def run_workflow(self,wf_name):
        """Execute the workflow indicated by `wf_name`"""
        wf = self.workflows[wf_name]
        self.message_callback('preparing workflow {} for execution'.format(wf_name))
        stk,diag = wf.execution_stack()
        wf_clone = self.prepare_wf(wf_name,stk)
        wf_clone.data_callback = wf.data_callback
        #wf_clone.message_callback = wf.message_callback
        self.wf_clones[wf_name] = wf_clone
        #if pool is None:
        self.wf_running[wf_name] = True
        wf_clone.run()
        self.message_callback('execution finished')
        #else:
        #    # TODO: how do we know when a cloned wf is finished? 
        #    # need to implement a finished_callback?
        #    # or is there something in the Process that can help?
        #    #wf_clone.wfFinished.connect( partial(self.stop_workflow,wf_name) )
        #    # temporary for debugging:
        #    wf_clone.run()
        #    self.message_callback('execution finished')
        #    #wf_proc = Process(target=wf_clone.run,name=wf_name)
        #    #pool[wf_name] = wf_proc 

    def stop_workflow(self,wf_name):
        """Stop the workflow indicated by `wf_name`"""
        self.message_callback('stopping workflow {}'.format(wf_name))
        if wf_name in self.wf_clones.keys():
            wf = self.wf_clones.pop(wf_name)
            wf_pgn_names = self.plugin_names.pop(wf_name)
            wf.stop()
            for pgn_name in wf_pgn_names:
                if not any(pgn_name in pnms for pnms in self.plugin_names.values()):
                    self.plugin_manager.stop_plugin(pgn_name)
        #else:
        #    wf = self.workflows[wf_name]
        #    wf.stop()
        self.wf_running[wf_name] = False

    def prepare_wf(self,wf_name,stk):
        """
        For all of the operations in stack stk,
        load all inputs that are not workflow items. 
        """ 
        wf = self.workflows[wf_name]
        wf_clone = wf.build_clone()
        plugin_names = wf.plugin_names 
        #for op_name,op in wf_clone.operations.items():
        #    op.message_callback = wf.message_callback
        #    op.data_callback = partial( wf.set_op_item,op_name )
        for lst in stk:
            for op_tag in lst:
                op = wf_clone.get_data_from_uri(op_tag)
                for inpname,il in op.input_locator.items():
                    if il.tp == pawstools.plugin_item:
                        # Plugins are expected to be safe to communicate from any thread,
                        # so they need not be cloned or copied
                        wf_clone.set_op_item(op_tag,'inputs.'+inpname,
                        self.get_plugin_data(il))
                    elif il.tp == pawstools.entire_workflow:
                        if il.val in self.workflows.keys():
                            wf_name = il.val
                            input_wf = self.workflows[il.val]
                            input_wf_stk,diag = input_wf.execution_stack()
                            new_wf = self.prepare_wf(wf_name,input_wf_stk)
                            #new_wf = wf.build_clone()
                            # NOTE: setting this data_callback causes segfaults.
                            # TODO: figure out why the segfaults,
                            # and think of a good way to relay data back to the original workflow
                            #new_wf.data_callback = self.workflows[wf_name].data_callback
                            #new_wf.message_callback = self.workflows[wf_name].message_callback
                            wf_clone.set_op_item(op_tag,'inputs.'+inpname,new_wf)
                    elif il.tp not in [pawstools.runtime_type,pawstools.workflow_item]:
                        # NOTE 1: runtime inputs should be set at runtime, 
                        # so that the input does not get serialized,
                        # when and if the workflow gets serialized
                        # NOTE 2: workflow_item inputs are retrieved during execution
                        wf_clone.set_op_item(op_tag,'inputs.'+inpname,
                        copy.deepcopy(il.val))
                for inpname,il in op.input_locator.items():
                    if il.tp == pawstools.plugin_item and il.val is not None:
                        vals = il.val
                        if not isinstance(vals,list):
                            vals = [vals]
                        for val in vals:
                            pgin_name = val.split('.')[0]
                            if not pgin_name in plugin_names:
                                plugin_names.append(pgin_name)
        self.plugin_names[wf_name] = plugin_names
        #for pgn_name in plugin_names:
        #    if not self.plugin_manager.plugin_running[pgn_name]:
        #        self.plugin_manager.start_plugin(pgn_name) 
        return wf_clone

    def get_plugin_data(self,il):
        if isinstance(il.val,list):
            return [self.plugin_manager.get_data_from_uri(v) for v in il.val]
        else:
            return self.plugin_manager.get_data_from_uri(il.val)

    def load_workflow(self,wf_name,wf_setup_dict):
        """Load a workflow from a dict that specifies its parameters.

        If `wf_name` is not unique, self.workflows[wf_name] is overwritten.

        Parameters
        ----------
        wf_name : str
            name to be given to the new workflow
        wf_setup_dict : dict
            dict specifying workflow setup
        """
        self.add_workflow(wf_name)
        wfins = wf_setup_dict.pop('WORKFLOW_INPUTS')
        wfouts = wf_setup_dict.pop('WORKFLOW_OUTPUTS')
        opflags = wf_setup_dict.pop('OP_ENABLE_FLAGS')
        for op_tag, op_setup_dict in wf_setup_dict.items():
            self.workflows[wf_name].load_operation(op_tag,op_setup_dict,self.op_manager)
            if not opflags[op_tag]:
                self.workflows[wf_name].disable_op(op_tag)
        for inpname,inpval in wfins.items():
            self.workflows[wf_name].connect_input(inpname,inpval)
        for outname,outval in wfouts.items():
            self.workflows[wf_name].connect_output(outname,outval)





import os
from collections import OrderedDict

import paws
from paws.core.workflows.WfManager import WfManager
from paws.core import pawstools


# specify workflow names: 
wf_names = ['read_files','main']
# specify operation names and corresponding modules: 
op_maps = OrderedDict.fromkeys(wf_names)
for wf_name in wf_names:
    op_maps[wf_name] = OrderedDict()
op_maps['main']['header_files'] = 'IO.FILESYSTEM.BuildFileList'
op_maps['main']['batch'] = 'EXECUTION.Batch'
#op_maps['main']['get_t_r0'] = 'PACKAGING.BATCH.XYDataFromBatch'
#op_maps['main']['write_t_r0'] = 'IO.CSV.WriteArrayCSV'

op_maps['read_files']['read_header'] = 'IO.BL15.ReadHeader_SSRL15'
op_maps['read_files']['time_temp'] = 'PACKAGING.BL15.TimeTempFromHeader'
op_maps['read_files']['populations_path'] = 'IO.FILESYSTEM.BuildFilePath'
op_maps['read_files']['read_populations'] = 'IO.YAML.LoadYAML'
op_maps['read_files']['params_path'] = 'IO.FILESYSTEM.BuildFilePath'
op_maps['read_files']['read_params'] = 'IO.YAML.LoadYAML'

#
#
#
#
op_maps['read_files']['write_params'] = 'IO.YAML.SaveYAML'
#
#
#
#

wfmgr = WfManager()
# add the workflows and activate/add the operations:
for wf_name,op_map in op_maps.items():
    wfmgr.add_workflow(wf_name)
    for op_name,op_uri in op_map.items():
        op = wfmgr.op_manager.get_operation(op_uri)
        wfmgr.workflows[wf_name].add_operation(op_name,op)

wf = wfmgr.workflows['main']

# input 0: header files location
wf.set_op_input('header_files','dir_path',None) 
wf.connect_input('header_dir','header_files.inputs.dir_path') 

# input 1: header file regex
wf.set_op_input('header_files','regex','*.txt') 
wf.connect_input('header_regex','header_files.inputs.regex') 

wf.connect_output('r0_vs_t','get_t_r0.outputs.x_y')

wf.set_op_input('batch','work_item','read_files','entire workflow')
wf.set_op_input('batch','input_arrays',['header_files.outputs.file_list'],'workflow item')
wf.set_op_input('batch','input_keys',['header_file'])

#wf.set_op_input('get_t_r0','batch_outputs','batch.outputs.batch_outputs','workflow item')
#wf.set_op_input('get_t_r0','x_key','time')
#wf.set_op_input('get_t_r0','y_key','r0_sphere')
#wf.set_op_input('get_t_r0','x_shift_flag',True)
#wf.set_op_input('get_t_r0','x_sort_flag',True)

#wf.set_op_input('write_t_r0','array','get_t_r0.outputs.x_y','workflow item')
#wf.set_op_input('write_t_r0','headers',['time (s)','r0_sphere (A)'])
#wf.set_op_input('write_t_r0','dir_path','header_files.inputs.dir_path','workflow item')
#wf.set_op_input('write_t_r0','filename','t_r0_sphere')

wf = wfmgr.workflows['read_files']

# input 0: header file
wf.set_op_input('read_header','file_path',None) 
wf.connect_input('header_file','read_header.inputs.file_path') 

# input 1: key for fetching temperature from header dictionary 
wf.set_op_input('time_temp','temp_key','TEMP')
wf.connect_input('temp_key','time_temp.inputs.temp_key')

wf.connect_output('filename','read_spectrum.outputs.filename')
wf.connect_output('time','time_temp.outputs.time')
wf.connect_output('temperature','time_temp.outputs.temperature')
wf.connect_output('populations','read_populations.outputs.yaml_output')
wf.connect_output('params','read_params.outputs.yaml_output')
wf.connect_output('r0_sphere','read_params.outputs.yaml_output.r0_sphere')

wf.set_op_input('time_temp','header_dict','read_header.outputs.header_dict','workflow item')
wf.set_op_input('time_temp','time_key','time')
wf.set_op_input('populations_path','dir_path','read_header.outputs.dir_path','workflow item')
wf.set_op_input('populations_path','filename','read_header.outputs.filename','workflow item')
wf.set_op_input('populations_path','suffix','_populations')
wf.set_op_input('populations_path','ext','yaml')
wf.set_op_input('read_populations','file_path','populations_path.outputs.file_path','workflow item')
wf.set_op_input('params_path','dir_path','read_header.outputs.dir_path','workflow item')
wf.set_op_input('params_path','filename','read_header.outputs.filename','workflow item')
wf.set_op_input('params_path','suffix','_saxs_params')
wf.set_op_input('params_path','ext','yaml')
wf.set_op_input('read_params','file_path','params_path.outputs.file_path','workflow item')

#
#
#
#
wf.set_op_input('write_params','file_path','read_params.inputs.file_path','workflow item')
wf.set_op_input('write_params','data','read_params.outputs.yaml_output','workflow item')
#
#
#
#

pawstools.save_to_wfl(os.path.join(pawstools.sourcedir,'core','workflows','SAXS','BL15','timeseries_plots.wfl'),wfmgr)




from collections import OrderedDict

import numpy as np

from ...Operation import Operation
        
inputs=OrderedDict(file_path=None)
outputs=OrderedDict(x=None,y=None,x_y=None)

class CSVToXYData(Operation):
    """
    Read a csv-formatted file as floats, 
    package into arrays of x values and y values.
    """

    def __init__(self):
        super(CSVToXYData, self).__init__(inputs, outputs)
        self.input_doc['file_path'] = 'path to .csv file'
        self.output_doc['x'] = 'numpy array built from first column of csv data'
        self.output_doc['y'] = 'numpy array built from second column of csv data'
        self.output_doc['x_y'] = 'n-by-2 numpy array with x and y values'

    def run(self):
        p = self.inputs['file_path']
        x = np.loadtxt(p, dtype=float, delimiter=',', usecols=[0])
        y = np.loadtxt(p, dtype=float, delimiter=',', usecols=[1])
        self.outputs['x'] = x
        self.outputs['y'] = y
        x_y = np.array([x,y]).T
        self.outputs['x_y'] = x_y 
        





from collections import OrderedDict
import time

from ..Operation import Operation

inputs = OrderedDict(duration=None)
outputs = OrderedDict()

class Sleep(Operation):
    """Operation for delaying the Workflow."""

    def __init__(self):
        super(Sleep,self).__init__(inputs,outputs) 
        self.input_doc['duration'] = 'sleep duration in seconds'
        
    def run(self):
        time.sleep(self.inputs['duration']) 




from collections import OrderedDict
import copy

from ..Operation import Operation

inputs = OrderedDict(data=None)
outputs = OrderedDict(data=None)

class Copy(Operation):
    """Copies one input with copy.deepcopy."""

    def __init__(self):
        super(Copy,self).__init__(inputs,outputs) 
        self.input_doc['data'] = 'any object'
        self.output_doc['data'] = 'the result of copy.deepcopy() on input `data`'
        
    def run(self):
        self.outputs['data'] = copy.deepcopy(self.inputs['data'])




from collections import OrderedDict

from ..Operation import Operation

from xrsdkit.tools.piftools import unpack_old_pif, make_pif 

inputs=OrderedDict(
    pif=None,
    flag=True)
outputs=OrderedDict(pif=None)

class PIFRefactor(Operation):
    """Build a new PIF from an old PIF"""

    def __init__(self):
        super(PIFRefactor,self).__init__(inputs,outputs)
        self.input_doc['pif'] = 'old pif object'
        self.output_doc['pif'] = 'new pif object'

    def run(self):
        if self.inputs['flag']:
            p = self.inputs['pif']
            expt_id, t_utc, q_I, T_C, feats, pops, params, rpt = unpack_old_pif(p)

            new_pops = {} 
            if bool(pops['unidentified']):
                new_pops['anonymous']=dict(structure='unidentified')
            else:
                if bool(pops['diffraction_peaks']): 
                    new_pops['superlattice']=dict(
                        structure='fcc',
                        settings={},
                        parameters={},
                        basis={})
                    new_pops['noise']=dict(
                        structure='diffuse',
                        parameters={},
                        basis={})
                    if bool(pops['guinier_porod']):
                        new_pops['precursor']=dict(
                            structure='diffuse',
                            parameters={},  
                            basis={'guinier_porod_precursor':{'guinier_porod':{}}}
                            )
                    if bool(pops['spherical_normal']):
                        new_pops['nanoparticles']=dict(
                            structure='diffuse',
                            parameters={},  
                            basis={'spherical_normal_nanoparticles':{'spherical_normal':{}}} 
                            )
                else:
                    new_pops['noise']=dict(
                        structure='diffuse',
                        parameters={'I0':params['I0_floor']},
                        basis={'flat_noise':dict(
                            flat={'amplitude':1}
                            )}
                        )
                    if bool(pops['guinier_porod']):
                        new_pops['precursor']=dict(
                            structure='diffuse',
                            parameters={'I0':1},
                            basis={'guinier_porod_precursor':dict(
                                guinier_porod={'G':params['G_gp'],'r_g':params['rg_gp'],'D':params['D_gp']}
                                )}
                            )
                    if bool(pops['spherical_normal']):
                        new_pops['nanoparticles']=dict(
                            structure='diffuse',
                            parameters={'I0':params['I0_sphere']},
                            basis={'spherical_nanoparticles':dict(
                                spherical_normal={'r0':params['r0_sphere'],'sigma':params['sigma_sphere']}
                                )}
                            )
    
            csys = make_pif(str(expt_id)+'_'+str(int(t_utc)),expt_id,t_utc,q_I,T_C,new_pops)
            self.outputs['pif'] = csys




from collections import OrderedDict
import os

import PIL 
import numpy as np

from ...Operation import Operation

inputs=OrderedDict(
    file_path=None,
    image_mode=None,
    image_size=None)
outputs=OrderedDict(
    image_data=None,
    PIL_Image=None,
    dir_path=None,
    filename=None)

class OpenRaw(Operation):
    """Get the data out of a .raw image"""

    def __init__(self):
        super(OpenRaw,self).__init__(inputs,outputs) 
        self.input_doc['file_path'] = 'path to a .raw image'
        self.input_doc['image_mode'] = 'PIL.Image mode- '\
            'see http://pillow.readthedocs.io'
        self.input_doc['image_size'] = 'two-element iterable indicating '\
            'the width and height (respectively) of the image'
        self.output_doc['image_data'] = '2D array representing pixel values '\
            'taken from the input file'
        self.output_doc['PIL_Image'] = 'The result of PIL.Image.frombytes()'
        self.output_doc['dir_path'] = 'Path to the directory the image came from'
        self.output_doc['filename'] = 'The image filename, no path, no extension'
        self.input_datatype['file_path'] = 'str'
        self.input_datatype['image_mode'] = 'str'
        self.input_datatype['image_size'] = 'list'
        
    def run(self):
        p = self.inputs['file_path']
        dir_path = os.path.split(p)[0]
        file_nopath = os.path.split(p)[1]
        file_noext = os.path.splitext(file_nopath)[0]
        self.outputs['dir_path'] = dir_path 
        self.outputs['filename'] = file_noext 
        md = self.inputs['image_mode']
        sz = self.inputs['image_size']
        pil_img = PIL.Image.frombytes(md,sz,open(p).read())
        self.outputs['PIL_Image'] = pil_img 
        self.outputs['image_data'] = np.array(pil_img)




from __future__ import print_function
import os
import glob
from collections import OrderedDict

import numpy as np

import paws
from paws.core import pawstools
from paws.core import plugins
from paws.core import operations
from paws.core import workflows 
from paws.core.plugins.PawsPlugin import PawsPlugin
from paws.core.workflows.WfManager import WfManager
from paws.core.workflows.Workflow import Workflow 
from paws.core.operations.Operation import Operation

def test_init():
    wf_manager = WfManager() 

def test_add_wf():
    wf_manager = WfManager() 
    wf_manager.add_workflow('test')
    assert isinstance(wf_manager.workflows['test'],Workflow)

def test_add_op():
    wf_manager = WfManager() 
    wf_test = wf_manager.add_workflow('test')
    test_listprimes = wf_manager.op_manager.get_operation('TESTS.ListPrimes')
    wf_test.add_operation('listprimes',test_listprimes)
    assert isinstance(test_listprimes,Operation)

def test_execute():
    wf_manager = WfManager() 
    wf_test = wf_manager.add_workflow('test')
    test_listprimes = wf_manager.op_manager.get_operation('TESTS.ListPrimes')
    wf_test.add_operation('listprimes',test_listprimes)
    wf_manager.run_workflow('test') 

def test_save():
    wf_manager = WfManager() 
    wf_test = wf_manager.add_workflow('test')
    test_listprimes = wf_manager.op_manager.get_operation('TESTS.ListPrimes')
    wf_test.add_operation('listprimes',test_listprimes)
    wfl_path = os.path.join(pawstools.paws_scratch_dir,'test.wfl')
    pawstools.save_to_wfl(wfl_path,wf_manager)

def test_load():
    wf_manager = WfManager()
    wfl_path = os.path.join(pawstools.paws_scratch_dir,'test.wfl') 
    pawstools.load_wfl(wfl_path,wf_manager)
    wf_manager.run_workflow('test') 

def test_get_output():
    wf_manager = WfManager()
    wfl_path = os.path.join(pawstools.paws_scratch_dir,'test.wfl') 
    pawstools.load_wfl(wfl_path,wf_manager)
    wf_manager.workflows['test'].connect_output('primes','listprimes.outputs.primes_list')
    wf_manager.run_workflow('test')
    p = wf_manager.workflows['test'].get_wf_output('primes')
    print(p)

def test_load_plugins():
    wf_manager = WfManager()
    for i,plugin_name in enumerate(plugins.plugin_name_list):
        print('loading {} ...'.format(plugin_name))
        pgin = wf_manager.plugin_manager.get_plugin(plugin_name)
        wf_manager.plugin_manager.add_plugin('plugin_'+str(i),plugin_name)
        assert isinstance(pgin,PawsPlugin)

def test_load_operations():
    wf_manager = WfManager()
    for op_module in wf_manager.op_manager.list_operations():
        print('loading {} ...'.format(op_module))
        op = wf_manager.op_manager.get_operation(op_module)
        assert isinstance(op,Operation)

def test_run_operations():
    wf_manager = WfManager()
    runnable_ops = dict.fromkeys(wf_manager.op_manager.list_operations())
    for op_module in wf_manager.op_manager.list_operations():
        print('running {} ...'.format(op_module))
        op = wf_manager.op_manager.get_operation(op_module)
        try:
            op.run()
        except Exception as ex:
            runnable_ops[op_module] = False
        runnable_ops[op_module] = True 
    return runnable_ops

def test_load_wfls():
    for wf_mod in workflows.wf_modules:
        wf_manager = WfManager()
        print('loading {} ...'.format(wf_mod))
        pawstools.load_packaged_wfl(wf_mod,wf_manager)




from __future__ import print_function
from collections import OrderedDict
import glob
import copy

import numpy as np

from ..Operation import Operation
from .. import optools

inputs=OrderedDict(
    work_item=None,
    input_arrays=[],
    input_keys=[],
    static_inputs=[],
    static_input_keys=[],
    pass_thru_params={},
    order_array=None,
    order_function=None,
    flag=True)

outputs=OrderedDict(
    batch_inputs=None,
    batch_outputs=None,
    flag=False)
        
class Batch(Operation):
    """Batch-execute a Workflow or Operation in specific order"""

    def __init__(self):
        super(Batch,self).__init__(inputs,outputs)
        self.input_doc['work_item'] = 'the Operation '\
            'or Workflow object to be batch-executed'

        self.input_doc['input_arrays'] = 'list of arrays, '\
            'one for each of the `input_keys`, '\
            'to be iterated over during batch execution'
        self.input_doc['input_keys'] = 'list of keys for setting '\
            'batch inputs of the `work_item`- these should correspond to '\
            'either Operation.inputs or Workflow.inputs, '\
            'depending on whether `work_item` is an Operation or a Workflow'

        self.input_doc['static_inputs'] = 'one object for each of '\
            'the `work_item` inputs indicated by `static_input_keys`, '\
            'to be set to the same value across the entire batch'
        self.input_doc['static_input_keys'] = 'list of keys for setting '\
            'static inputs of the `work_item`- these should correspond to '\
            'either Operation.inputs or Workflow.inputs, '\
            'depending on whether `work_item` is an Operation or a Workflow'

        self.input_doc['pass_thru_params'] = 'dict of '\
            'input_name:output_name pairs, where the `work_item` input '\
            'for input_name is set using the `work_item` output output_name '\
            'from the previous execution of `work_item`. '\
            'This step is skipped for the first `work_item` in the batch. '

        self.input_doc['order_array'] = '(optional) array, '\
            'similar to input_arrays, used in conjunction '\
            'with `order_function` to determine '\
            'the order of execution of the batch- '\
            'if `order_function` is not provided, '\
            '`order_array` should be an array of real numbers'
        self.input_doc['order_function'] = '(optional) function '\
            'called on elements of `order_array` to determine '\
            'execution order, which will be '\
            'in increasing order of the return value'
        self.input_doc['flag'] = 'Flag for whether or not to run the batch' 

        self.output_doc['batch_inputs'] = 'list of dicts, '\
            'where each dict gives the input_key:input_value pairs '\
            'used in each execution of `work_item`, '\
            'in batch execution order'
        self.output_doc['batch_outputs'] = 'list of dicts, '\
            'where each dict presents the `work_item` outputs, '\
            'in batch execution order'
        self.output_doc['flag'] = 'Flag for whether or not the batch completed' 

        self.input_datatype['input_arrays'] = 'list'
        self.input_datatype['input_keys'] = 'list'
        self.input_datatype['static_inputs'] = 'list'
        self.input_datatype['static_input_keys'] = 'list'
        self.input_datatype['pass_thru_params'] = 'dict'

    def run(self):
        wrkitm = self.inputs['work_item']

        inps = self.inputs['input_arrays']
        inpks = self.inputs['input_keys']
        stat_inps = self.inputs['static_inputs']
        stat_inpks = self.inputs['static_input_keys']
        pass_thru_params = self.inputs['pass_thru_params']
        f = self.inputs['flag']

        if bool(f):
            # index the execution order
            odrvals = self.inputs['order_array']
            odrfnc = self.inputs['order_function']
            if odrfnc is not None and odrvals is not None:
                odrvals = np.array([odrfnc(v) for v in odrvals])
                odr_idx = np.argsort(odrvals)
            else:
                odr_idx = np.arange(len(inps[0]))
            n_batch = len(odr_idx)

            #wrkitms = [wrk.build_clone() for idx in odr_idx]
            #wrkitms = [wrk for idx in odr_idx]
            
            batch_inp_dicts = []
            for idx in odr_idx:
                batch_inps = OrderedDict.fromkeys(inpks)
                for inpk,inp in zip(inpks,inps):
                    #wrkitm.set_input(inpk,inp[idx])
                    batch_inps[inpk] = inp[idx]
                batch_inp_dicts.append(batch_inps)

            self.outputs['batch_inputs'] = batch_inp_dicts 
            self.outputs['batch_outputs'] = [None for ib in range(n_batch)] 
            if self.data_callback: 
                self.data_callback('outputs.batch_inputs',copy.deepcopy(batch_inp_dicts))
                self.data_callback('outputs.batch_outputs',[None for ib in range(n_batch)])

            self.message_callback('STARTING BATCH')
            out_dict = None
            for i,idx in enumerate(odr_idx): 

                if self.stop_flag:
                    self.message_callback('Batch stopped.')
                    return
               
                inp_dict = batch_inp_dicts[idx]
                for inp_name,inp_value in inp_dict.items():
                    wrkitm.set_input(inp_name,inp_value)
 
                self.message_callback('BATCH RUN {} / {}'.format(i,n_batch-1))
                #wrki = wrkitms[idx]

                if any(stat_inpks): 
                    for inpnm,inpval in zip(stat_inpks,stat_inps):
                        wrkitm.set_input(inpnm,inpval)

                if any(pass_thru_params) and out_dict is not None:
                    for inp_name,out_name in pass_thru_params.items():
                        wrkitm.set_input(inp_name,out_dict[out_name])

                wrkitm.run()
                out_dict = copy.deepcopy(wrkitm.get_outputs())
                self.outputs['batch_outputs'][i] = out_dict
                if self.data_callback: 
                    self.data_callback('outputs.batch_outputs.'+str(i),copy.deepcopy(out_dict))
            self.outputs['flag'] = True
            self.message_callback('BATCH FINISHED')
        else:
            self.outputs['flag'] = False 
            self.message_callback('BATCH SKIPPED')



from collections import OrderedDict
import os

from pypif import pif
import pypif.obj as pifobj 

from ...Operation import Operation

inputs=OrderedDict(
    pif=None,
    dir_path=None,
    filename=None)
outputs=OrderedDict(file_path=None)
        
class SavePIFAsJSON(Operation):

    def __init__(self):
        super(SavePIFAsJSON,self).__init__(inputs,outputs)
        self.input_doc['pif'] = 'A pypif.obj.System object or an array/list thereof'
        self.input_doc['dir_path'] = 'Path to system directory where .json file will be created' 
        self.input_doc['filename'] = 'Name of the .json file to be created. '\
            'The .json extension is appended automatically if not provided.' 
        self.output_doc['file_path'] = 'Full path to the newly saved .json file'

    def run(self):
        p = self.inputs['pif']        
        dp = self.inputs['dir_path']
        fn = self.inputs['filename']
        if not os.path.splitext(fn)[1] == 'json':
            fn = fn+'.json'
        json_file = os.path.join(dp,fn)
        self.outputs['file_path'] = json_file
        pif.dump(p, open(json_file,'w'))




from collections import OrderedDict

import numpy as np

import yaml
from ...Operation import Operation

inputs=OrderedDict(
    file_path=None,
    data=None)
outputs=OrderedDict(
    yaml_output=None)

class SaveYAML(Operation):
    """Save some data to a YAML file."""

    def __init__(self):
        super(SaveYAML, self).__init__(inputs, outputs)
        self.input_doc['file_path'] = 'path where YAML file will be saved'
        self.input_doc['data'] = 'data to save in the YAML file' 

    def run(self):
        p = self.inputs['file_path']
        d = self.inputs['data']

        stream = open(p,'w')
        yaml.dump(d, stream)

        #datastring = yaml.dump(data)
        #self.outputs['data_string'] = datastring[:min([len(datastring),1000])]




import os
from collections import OrderedDict

import paws
from paws.core.workflows.WfManager import WfManager
from paws.core import pawstools

# specify workflow names: 
wf_names = ['saxs_gp_fit','batch_saxs_gp_fit']
# specify operation names and corresponding modules: 
op_maps = OrderedDict.fromkeys(wf_names)
for wf_name in wf_names:
    op_maps[wf_name] = OrderedDict()
op_maps['batch_saxs_gp_fit']['Batch Execution'] = 'EXECUTION.Batch'
op_maps['saxs_gp_fit']['Read Data'] = 'IO.NUMPY.Loadtxt_q_I_dI'
op_maps['saxs_gp_fit']['Q-Window'] = 'PACKAGING.Window'
op_maps['saxs_gp_fit']['Log(I)'] = 'PROCESSING.BASIC.LogY'
op_maps['saxs_gp_fit']['Fit Spectrum'] = 'PROCESSING.FITTING.XRSDFit'
op_maps['saxs_gp_fit']['Log(I_fit)'] = 'PROCESSING.BASIC.LogY'
op_maps['saxs_gp_fit']['Output CSV'] = 'IO.CSV.WriteArrayCSV'

wfmgr = WfManager()
# add the workflows and activate/add the operations:
for wf_name,op_map in op_maps.items():
    wfmgr.add_workflow(wf_name)
    for op_name,op_uri in op_map.items():
        op = wfmgr.op_manager.get_operation(op_uri)
        wfmgr.workflows[wf_name].add_operation(op_name,op)

# define guinier-porod and noise populations
pops = {} 
pops['noise'] = dict(
    structure = 'diffuse',
    parameters = {'I0':0.1},
    basis = {'flat_noise':{'flat':{}}}
    )
pops['scatterers'] = dict(
    structure = 'diffuse',
    parameters = {},
    basis = {'gp_scatterers':{'guinier_porod':{'G':100.,'rg':10.,'D':4.}}}
    )

wf = wfmgr.workflows['saxs_gp_fit']

wf.connect_input('file_path','Read Data.inputs.file_path')
wf.set_op_input('Read Data','file_path',None)

wf.set_op_input('Read Data','delimiter',',')
wf.set_op_input('Q-Window','x_y','Read Data.outputs.q_I','workflow item')
wf.set_op_input('Q-Window','x_min',0.02)
wf.set_op_input('Q-Window','x_max',0.5)
wf.set_op_input('Log(I)','x_y','Q-Window.outputs.x_y_window','workflow item')
wf.set_op_input('Fit Spectrum','q_I','Q-Window.outputs.x_y_window','workflow item')
wf.set_op_input('Fit Spectrum','populations',pops)
wf.set_op_input('Log(I_fit)','x_y','Fit Spectrum.outputs.q_I_opt','workflow item')
wf.set_op_input('Output CSV','array','Fit Spectrum.outputs.q_I_opt','workflow item')
wf.set_op_input('Output CSV','headers',['q','I'])
wf.set_op_input('Output CSV','dir_path','Read Data.outputs.dir_path','workflow item')
wf.set_op_input('Output CSV','filename','Read Data.outputs.filename','workflow item')
wf.set_op_input('Output CSV','filetag','_fit')

wf.disable_op('Output CSV')

wf = wfmgr.workflows['batch_saxs_gp_fit']

wf.set_op_input('Batch Execution','work_item','saxs_gp_fit','entire workflow')
wf.set_op_input('Batch Execution','input_arrays',None)
wf.set_op_input('Batch Execution','input_keys',['file_path'])

pawstools.save_to_wfl(os.path.join(pawstools.sourcedir,'core','workflows','XICAM','batch_saxs_gp_fit.wfl'),wfmgr)




from collections import OrderedDict
import os

from pypif import pif
import pypif.obj as pifobj 

from ...Operation import Operation

inputs=OrderedDict(
    pif=None,
    client_plugin=None,
    dsid=None,
    json_dirpath=None,
    json_filename=None,
    keep_json=False,
    ship_flag=False)
outputs=OrderedDict(response=None)
        
class ShipToDataSet(Operation):
    """
    Take a pypif.obj.System object and ship it to a given Citrination data set.    
    """

    def __init__(self):
        super(ShipToDataSet,self).__init__(inputs,outputs)
        self.input_doc['pif'] = 'A pypif.obj.System object or an array/list thereof'
        self.input_doc['client_plugin'] = 'A running CitrinationClient(PawsPlugin)' 
        self.input_doc['dsid'] = 'Data set ID where the pif record(s) will be stored on Citrination' 
        self.input_doc['json_dirpath'] = 'Filesystem path where a json file of the pif(s) will be saved' 
        self.input_doc['json_filename'] = 'Name of the .json file where the pif(s) will be saved' 
        self.input_doc['keep_json'] = 'Flag for whether or not to keep the json file' 
        self.input_doc['ship_flag'] = 'Flag for shipping the pif- set to False for a dry run' 
        self.output_doc['response'] = 'The Citrination server response to the shipment'

    def run(self):
        cl_pgn = self.inputs['client_plugin'] 
        dsid = self.inputs['dsid'] 
        p = self.inputs['pif']        
        json_dir = self.inputs['json_dirpath']
        json_file = self.inputs['json_filename']
        if not os.path.splitext(json_file)[1] == 'json':
            json_file = json_file+'.json'
        json_file = os.path.join(json_dir,json_file)

        json_flag = self.inputs['keep_json']
        ship_flag = self.inputs['ship_flag']
        self.message_callback('dumping PIF data to {}'.format(json_file))
        pif.dump(p, open(json_file,'w'))
        if ship_flag:
            self.message_callback('Uploading {} to dataset {}'.format(json_file,dsid))
            r = cl_pgn.client.upload_file(json_file,dataset_id = dsid)
        else:
            r = 'dry run: no shipment occurred.'
        if not json_flag:
            self.message_callback('Removing {}'.format(json_file))
            os.remove(json_file) 
        self.outputs['response'] = r





from collections import OrderedDict

from ...Operation import Operation

inputs=OrderedDict(
    ssh_file_copier=None,
    remote_dir=None,
    regex='*',
    local_dir=None)
outputs=OrderedDict(
    files_copied=[])

class SSHCopy(Operation):
    """Use SSHFileCopier Plugin to get files from remote filesystem"""

    def __init__(self):
        super(SSHCopy, self).__init__(inputs, outputs)
        self.input_doc['ssh_file_copier'] = 'SSHFileCopier plugin'
        self.input_doc['remote_dir'] = 'remote directory path (file source)'
        self.input_doc['regex'] = 'regular expression for selecting remote files'
        self.input_doc['local_dir'] = 'local directory path (file destination)'
        self.output_doc['files_copied'] = 'full (local) paths to files that were requested' 

    def run(self):
        rdir = self.inputs['remote_dir']
        rx = self.inputs['regex']
        ldir = self.inputs['local_dir']
        self.inputs['ssh_file_copier'].request_copy(rdir,rx,ldir)




import os
from collections import OrderedDict

import paws
from paws.core.workflows.WfManager import WfManager
from paws.core import pawstools

# specify workflow names: 
wf_names = ['saxs_integrate']
# specify operation names and corresponding modules: 
op_map = OrderedDict()
op_map['read_calibration'] = 'IO.CALIBRATION.ReadPONI'
op_map['build_integrator'] = 'PROCESSING.INTEGRATION.BuildPyFAIIntegrator'
op_map['read_image'] = 'IO.IMAGE.FabIOOpen'
op_map['integrate_to_1d'] = 'PROCESSING.INTEGRATION.ApplyIntegrator1d'
op_map['q_window'] = 'PACKAGING.Window'
op_map['dezinger'] = 'PROCESSING.ZINGERS.EasyZingers1d'
op_map['log_I'] = 'PROCESSING.BASIC.LogY'
op_map['output_CSV'] = 'IO.CSV.WriteArrayCSV'

wfmgr = WfManager()
wfmgr.add_workflow('saxs_integrate')
wf = wfmgr.workflows['saxs_integrate']
for op_name,op_uri in op_map.items():
    op = wfmgr.op_manager.get_operation(op_uri)
    wf.add_operation(op_name,op)

# input 1: path to .nika calibration file
wf.set_op_input('read_calibration','file_path','')
wf.connect_input('poni_file','read_calibration.inputs.file_path')
# input 2: path to image file
wf.set_op_input('read_image','file_path','')
wf.connect_input('file_path','read_image.inputs.file_path')
# inputs 3 and 4: q-range for data windowing
wf.set_op_input('q_window','x_min',0.0)
wf.set_op_input('q_window','x_max',1.0)
wf.connect_input('q_min','q_window.inputs.x_min')
wf.connect_input('q_max','q_window.inputs.x_max')
# input 5: path to output directory
wf.set_op_input('output_CSV','dir_path',os.getcwd())
wf.connect_input('output_dir','output_CSV.inputs.dir_path')

### (4) set up the rest of the Operation IO routing
wf.set_op_input('build_integrator','poni_dict','read_calibration.outputs.poni_dict','workflow item')
wf.set_op_input('integrate_to_1d','image_data','read_image.outputs.image_data','workflow item')
wf.set_op_input('integrate_to_1d','integrator','build_integrator.outputs.integrator','workflow item')
wf.set_op_input('q_window','x_y','integrate_to_1d.outputs.q_I','workflow item')
wf.set_op_input('dezinger','q_I','q_window.outputs.x_y_window','workflow item')
wf.set_op_input('log_I','x_y','dezinger.outputs.q_I_dz','workflow item')
wf.set_op_input('output_CSV','array','dezinger.outputs.q_I_dz','workflow item')
wf.set_op_input('output_CSV','headers',['q (1/angstrom)','intensity (arb)'])
wf.set_op_input('output_CSV','filename','read_image.outputs.filename','workflow item')
wf.set_op_input('output_CSV','filetag','_dz')

pawstools.save_to_wfl(os.path.join(pawstools.sourcedir,'core','workflows','SAXS','saxs_integrate.wfl'),wfmgr)




from collections import OrderedDict

import numpy as np

from ..Operation import Operation

inputs=OrderedDict(
    x_y=None,
    x_min=None,
    x_max=None)
outputs=OrderedDict(x_y_window=None)

class Window(Operation):
    """
    Window an n-by-2 array x_y 
    such that x is bounded by specified limits 
    """

    def __init__(self):
        super(Window,self).__init__(inputs,outputs)        
        self.input_doc['x_y'] = 'n-by-2 array of x and y values'
        self.input_doc['x_min'] = 'inclusive minimum x value of output'
        self.input_doc['x_max'] = 'inclusive maximum x value of output'
        self.output_doc['x_y_window'] = 'n-by-2 array with x, y pairs for x_min <= x <= x_max'
        self.input_datatype['x_min'] = 'float'
        self.input_datatype['x_max'] = 'float'

    def run(self):
        x_y = self.inputs['x_y']
        x_min = self.inputs['x_min']
        x_max = self.inputs['x_max']
        idx_keep = ((x_y[:,0] >= x_min) & (x_y[:,0] <= x_max))
        self.outputs['x_y_window'] = x_y[idx_keep,:] 




from __future__ import print_function
from collections import OrderedDict
import serial
import copy
import os
import sys
from threading import Thread, Condition
if int(sys.version[0]) == 2:
    import Queue as queue
else:
    import queue 

from .PawsPlugin import PawsPlugin
from .. import pawstools

inputs = OrderedDict(
    serial_device=None,
    timer=None,
    verbose=False)

class MitosPPumpController(PawsPlugin):
    """PAWS Plugin for controlling a Mitos P-pump.

    Uses a serial port (or virtual serial port)
    to communicate to a USB-attached Mitos P-pump.
    RS232 protocol settings: 
    57600 baud, 8 data bits, 1 stop bit, no parity, no handshaking.
    The pump controller acts as a proxy for controlling a clone of itself.
    The clone operates in its own thread, exchanging information 
    with the original pump controller in a threadsafe way,
    at each tick of the input timer.

    When the pump is in CONTROL mode,
    a command must be sent every 30 seconds,
    or the pump will automatically exit CONTROL mode.
    """
    def __init__(self):
        super(MitosPPumpController,self).__init__(inputs)
        self.input_doc['serial_device'] = 'string serial device indicator '\
            '(e.g. "COM2") or filesystem path (e.g. "/dev/ttyUSB0")'
        self.input_doc['timer'] = 'Timer PawsPlugin, used to initiate '\
            'thread-safe data exchanges between the pump controller '\
            'and its operational clone.'
        self.ErrorsDict = { 0:'None', \
                            1:'Supply than maximum', \
                            2:'Tare time out', \
                            3:'Tare supply still connected ', \
                            4:'Control start time out', \
                            5:'Pressure target too low', \
                            6:'Pressure target too high', \
                            7:'Leak test supply pressure too low', \
                            8:'Leak test time out', \
                            100:'Broken'}
        self.StateDict  = { 0:'IDLE', \
                            1:'CONTROLLING', \
                            2:'TARE', \
                            3:'ERROR', \
                            4:'LEAK TEST'}
        self.ser = None
        self.state = OrderedDict(
            error_code = None,
            state_code = None, 
            chamber_pressure = None, 
            supply_pressure = None, 
            target_pressure = None, 
            flow_rate = None, 
            target_flow_rate = None)
        # a lock for pump history 
        self.history_lock = Condition()
        self.history = []
        # a lock for pump controls
        #self.control_lock = Condition()
        # a lock for signalling run/stop
        self.running_lock = Condition()
        #self.n_events = 0
        self.commands = queue.Queue() 
        self.thread_clone = None
        self.proxy = None

    def start(self):
        with self.running_lock:
            self.running = True
        self.thread_clone = self.build_clone()
        self.thread_clone.proxy = self
        self.thread_clone.ser = serial.Serial(
            self.inputs['serial_device'], 
            57600, timeout=1, 
            parity = serial.PARITY_NONE, 
            bytesize = serial.EIGHTBITS, 
            xonxoff = 0, rtscts = 0)
        th = Thread(target=self.thread_clone.run)
        th.start()
       
    def run(self):
        vb = self.inputs['verbose']
        # this method is run by self.thread_clone 
        keep_going = True
        # check for control...
        self.read_status()
        if not self.state['state_code'] == 1:
            # attempt to enter remote control mode 
            cmd = "A1"
            self.send_line(cmd)  
            resp = self.receive_line()
            with self.proxy.history_lock:
                with self.proxy.inputs['timer'].dt_lock:
                    t_now = float(self.proxy.inputs['timer'].time_points[-1])
                self.proxy.add_to_history(t_now,str(cmd),str(resp))
            if not resp.decode() == "#A0":
                msg = "Pump failed to enter REMOTE control mode"
                with self.proxy.history_lock:
                    with self.proxy.inputs['timer'].dt_lock:
                        t_now = float(self.proxy.inputs['timer'].time_points[-1])
                    self.proxy.add_to_history(t_now,'ERROR',str(msg))
                keep_going = False
                with self.proxy.running_lock:
                    self.proxy.stop()
                #raise RuntimeError(msg)

        # clear the pump...
        cmd = "C"
        self.send_line(cmd)
        resp = self.receive_line()
        with self.proxy.history_lock:
            with self.proxy.inputs['timer'].dt_lock:
                t_now = float(self.proxy.inputs['timer'].time_points[-1])
            self.proxy.add_to_history(t_now,str(cmd),str(resp))

        while keep_going:
            with self.proxy.inputs['timer'].dt_lock:
                self.proxy.inputs['timer'].dt_lock.wait()
            self.read_status()
            if vb: self.message_callback(self.print_status())
            while self.commands.qsize() > 0: 
                cmd = str(self.commands.get())
                self.commands.task_done()
                self.send_line(cmd)
                resp = self.receive_line()
                with self.proxy.history_lock:
                    with self.proxy.inputs['timer'].dt_lock:
                        t_now = float(self.proxy.inputs['timer'].time_points[-1])
                    self.proxy.add_to_history(t_now,str(cmd),str(resp))
            with self.proxy.inputs['timer'].running_lock:
                if not self.proxy.inputs['timer'].running:
                    with self.proxy.running_lock:
                        self.proxy.stop()
            with self.proxy.running_lock:
                keep_going = bool(self.proxy.running)
        cmd = "A0"
        with self.proxy.inputs['timer'].dt_lock:
            t_now = float(self.proxy.inputs['timer'].time_points[-1])
        self.send_line(cmd)
        resp = self.receive_line()
        with self.proxy.history_lock:
            self.proxy.add_to_history(t_now,str(cmd),str(resp))
            self.proxy.dump_history()
        self.ser.close()

    def send_line(self,line):
        self.ser.write("{}\r\n".format(line).encode('utf-8'))  

    def receive_line(self):
        r = self.ser.readline()
        return r.strip()

    def add_to_history(self,t,cmd,resp):
        self.history.append(self.event(t,cmd,resp))
        #print('\nMitosPPumpController command: {}'.format(cmd))
        #print('response: {}'.format(resp))

    def event(self,t,cmd,resp):
        event = OrderedDict(t=t,command=cmd,response=resp)
        return event

    def dump_history(self):
        dump_path = os.path.join(pawstools.paws_scratch_dir,'ppump_controller_{}.log'.format(self))
        dump_file = open(dump_path,'w')
        dump_file.write('t :: command :: response\n')
        for ev in self.history:
            dump_file.write('{} :: {} :: {}\n'.format(ev['t'],ev['command'],ev['response']))
        dump_file.close()


    def read_status(self):
        """ Read pump status.

        The reply from the pump is formatted as 
        #sErr,Sp,Src,Pc,Ps,Pt,Qc,Qt,Ft
        These are numeric codes defining the response state.  

        Err - Error code. A value of 0 means CMD_ACCEPTED - OK.

        Sp  - State of the pump
            0.   IDLE
            1.   CONTROLLING
            2.   TARE
            3.   ERROR
            4.   LEAK TEST
        The status returned by the pump should reflect
        the latest command sent from the PC to the pump.

        Src - Mode of control: 0 is manual, 1 is remote controlled.

        Pc  - Chamber pressure in mbar.

        Ps  - Supply pressure in mbar.

        Pt  - Target pressure in mbar. 

        Qc  - Current flow rate in pl/s. 
            Requires a connected Mitos Sensor Display flow meter.

        Qt  - Target flow rate in pl/s. 
            Requires a connected Mitos Sensor Display flow meter.

        Ft  - Flow sensor type plus flow control mode. 
            Requires a connected Mitos Sensor Display,
            or sensor interface flow meter.
            This consists of an upper nibble, 
            middle nibble and lower nibble.
        """
        cmd = "s"
        self.send_line(cmd)
        resp = self.receive_line()
        s = resp.split(b',')
        while '' in s or len(s) < 8:
            self.send_line(cmd)
            resp = self.receive_line()
            s = resp.split(b',')
        with self.proxy.history_lock:
            with self.proxy.inputs['timer'].dt_lock:
                t_now = float(self.proxy.inputs['timer'].time_points[-1])
            self.proxy.add_to_history(t_now,str(cmd),str(resp))
        self.state['error_code'] = int(s[0][2:])
        self.state['state_code'] = int(s[1])
        self.state['chamber_pressure'] = float(s[3])
        self.state['supply_pressure'] = float(s[4])
        self.state['target_pressure'] = float(s[5])
        self.state['flow_rate'] = float(s[6])
        self.state['target_flow_rate'] = float(s[7])
        # TODO: perform this update using the queue
        #if self.data_callback:
        #    self.data_callback('content.state',copy.deepcopy(self.state))

    def print_status(self):
        msg = '----------- PUMP STATUS -----------'+os.linesep
        msg += 'Error code: {} ({})'.format(self.state['error_code'],
            self.ErrorsDict[self.state['error_code']])+os.linesep
        msg += 'State code: {} ({})'.format(self.state['state_code'],
            self.StateDict[self.state['state_code']])+os.linesep
        msg += 'Chamber pressure: \t{}'.format(self.state['chamber_pressure'])+os.linesep 
        msg += 'Supply pressure: \t{}'.format(self.state['supply_pressure'])+os.linesep 
        msg += 'Target pressure: \t{}'.format(self.state['target_pressure'])+os.linesep 
        msg += 'Current flow rate: \t{}'.format(self.state['flow_rate'])+os.linesep 
        msg += 'Target flow rate: \t{}'.format(self.state['target_flow_rate'])+os.linesep 
        msg += '-----------------------------------'
        return msg

    def set_flowrate(self,rate):
        """Set the pump flow rate to `rate` (microlitres per minute).

        Parameters
        ----------
        rate : integer
            The flow rate set point, in microlitres per minute.
            The pump controller expects and integer value in
            picolitres per second.
            The rate is rounded to the nearest integer value
            after converting it to picolitres per second.
        """
        pl_s_rate = int(round(float(rate)*1.E6/60.))
        self.thread_clone.commands.put("F{}".format(pl_s_rate)) 

    def set_pressure(self,pressure):
        """Set the pump pressure to the provided value.

        Parameters
        ----------
        pressure : integer
            The pressure set point, in mbar.
            A value of 1000 sends the 'P1000' command,
            which sets target pressure to 1000 mbar. 
        """
        self.thread_clone.commands.put("P{}".format(pressure)) 

    def tare(self):
        """Tare the P-pump."""
        if self.inputs['verbose']: self.message_callback('taring pump')
        self.thread_clone.commands.put("R0")

    def set_idle(self):
        """Set the P-pump to idle."""
        if self.inputs['verbose']: self.message_callback('setting pump to idle')
        self.thread_clone.commands.put("P0")

    def dispense_volume(self, vol):
        """Control the pump to dispense a specified volume.

        Parameters
        ----------
        vol : float
            The volume to dispense 
        """
        pass

    def get_plugin_content(self):
        with self.history_lock:
            h = copy.deepcopy(self.history)
        return {'history':h}





from __future__ import print_function
from collections import OrderedDict
import copy
import socket 
import os
import sys
from threading import Thread, Condition
if int(sys.version[0]) == 2:
    import Queue as queue
else:
    import queue 

import tzlocal

from .PawsPlugin import PawsPlugin
from .. import pawstools

inputs = OrderedDict(
    host=None,
    port=None,
    timer=None,
    verbose=False)

class SpecInfoClient(PawsPlugin):

    def __init__(self):
        super(SpecInfoClient,self).__init__(inputs)
        self.input_doc['host'] = 'string representing host name or IP address'
        self.input_doc['port'] = 'integer port number where SpecInfoServer listens' 
        self.input_doc['timer'] = 'Timer plugin for triggering client activities' 
        self.input_doc['verbose'] = 'If True, plugin uses its message_callback' 
        self.n_events = 0
        self.history_lock = Condition()
        self.history = []
        self.commands = queue.Queue() 
        self.running_lock = Condition()
        self.sock = None
        self.content = OrderedDict(history = self.history)
        self.thread_clone = None
        self.proxy = None

    def description(self):
        desc = 'SpecInfoClient Plugin: '\
            'This is a TCP Client used to communicate with SpecInfoServer. '\
            'Startup requires a host name and a port number, '\
            'where SpecInfoServer should be listening.'
        return desc

    def start(self):
        with self.running_lock:
            self.running = True
        hst = self.inputs['host'] 
        prt = self.inputs['port'] 
        tmr = self.inputs['timer'] 
        self.thread_clone = self.build_clone()
        self.thread_clone.proxy = self
        self.thread_clone.sock = socket.create_connection((hst,prt)) 
        th = Thread(target=self.thread_clone.run)
        th.start()

    def run(self):
        vb = self.inputs['verbose']
        # executed by self.thread_clone in its own thread
        cmd = '!rqc'
        with self.proxy.inputs['timer'].dt_lock:
            t_now = float(self.proxy.inputs['timer'].time_points[-1])
        self.send_line(cmd)
        resp = self.receive_line()
        if vb: self.message_callback('{} :: {} :: {}'.format(t_now,cmd,resp.decode()))
        with self.proxy.history_lock:
            self.proxy.add_to_history(t_now,cmd,resp.decode())
        while not resp.decode() == 'client in control.':
            with self.proxy.inputs['timer'].dt_lock:
                self.proxy.inputs['timer'].dt_lock.wait()
            with self.proxy.inputs['timer'].dt_lock:
                t_now = float(self.proxy.inputs['timer'].time_points[-1])
            self.send_line(cmd)
            resp = self.receive_line()
            if vb: self.message_callback('{} :: {} :: {}'.format(t_now,cmd,resp.decode()))
            with self.proxy.history_lock:
                self.proxy.add_to_history(t_now,cmd,resp.decode())
        #    trial += 1
        #    if trial > 100:
        #        self.proxy.add_to_history('CONTROL DENIED',resp)
        #        raise RuntimeError('SpecInfoServer control denied')
        #if vb: self.message_callback('{} :: {} :: {}'.format(t_now,cmd,resp))
        keep_going = True
        while keep_going: 
            while self.commands.qsize() > 0:
                cmd = self.commands.get()
                # release self.commands with a task_done()
                self.commands.task_done()
                with self.proxy.inputs['timer'].dt_lock:
                    t_now = float(self.proxy.inputs['timer'].time_points[-1])
                self.send_line(cmd)
                resp = self.receive_line()
                if vb: self.message_callback('{} :: {} :: {}'.format(t_now,cmd,resp.decode()))
                while resp.decode() in ['','spec is busy!']:
                    with self.proxy.inputs['timer'].dt_lock:
                        self.proxy.inputs['timer'].dt_lock.wait()
                    with self.proxy.inputs['timer'].dt_lock:
                        t_now = float(self.proxy.inputs['timer'].time_points[-1])
                    self.send_line(cmd)
                    resp = self.receive_line()
                    if vb: self.message_callback('{} :: {} :: {}'.format(t_now,cmd,resp.decode()))
                with self.proxy.history_lock:
                    self.proxy.add_to_history(t_now,cmd,resp.decode())
                #with self.proxy.inputs['timer'].dt_lock:
                #    self.proxy.inputs['timer'].dt_lock.wait()
            with self.proxy.inputs['timer'].dt_lock:
                self.proxy.inputs['timer'].dt_lock.wait()
            with self.proxy.inputs['timer'].running_lock:
                if not self.proxy.inputs['timer'].running:
                    with self.proxy.running_lock:
                        self.proxy.stop()
            with self.proxy.running_lock:
                keep_going = bool(self.proxy.running)
        with self.proxy.inputs['timer'].dt_lock:
            t_now = float(self.proxy.inputs['timer'].time_points[-1])
        with self.proxy.history_lock:
            self.proxy.add_to_history(t_now,cmd,resp.decode())
            self.proxy.dump_history()
        self.sock.close()

    def dump_history(self):
        dump_path = os.path.join(pawstools.paws_scratch_dir,'ppump_controller_{}.log'.format(self))
        dump_file = open(dump_path,'w')
        dump_file.write('t :: command :: response\n')
        for ev in self.history:
            dump_file.write('{} :: {} :: {}\n'.format(ev['t'],ev['command'],ev['response']))
        dump_file.close()

    def get_plugin_content(self):
        with self.history_lock:
            h = copy.deepcopy(self.history)
        return {'history':h}

    def add_to_history(self,t,cmd,resp):
        self.history.append(self.event(t,cmd,resp))

    def event(self,t,cmd,resp):
        event = OrderedDict(t=t,command=cmd,response=resp)
        return event

    def receive_line(self):
        bfr = bytearray(b' ' * 1024) 
        ln = self.sock.recv_into(bfr) 
        bfr = bfr.strip()
        return bfr
    
    def send_line(self, line):
        self.sock.sendall(bytearray(line.encode('utf-8')))

    def enable_cryocon(self):
        self.thread_clone.commands.put('!cmd ctemp_enable')
        self.thread_clone.commands.put('!cmd ctemp_ctrl_on')
        self.thread_clone.commands.put('!cmd csettemp 20')

    def set_cryocon(self,temperature,ramp=None):
        if ramp is not None:
            self.thread_clone.commands.put('!cmd ctemp_ramp_on')
            self.thread_clone.commands.put('!cmd csetramp {}'.format(ramp))
        self.thread_clone.commands.put('!cmd csettemp {}'.format(temperature))

    def stop_cryocon(self):
        self.thread_clone.commands.put('!cmd ctemp_ctrl_off')
        self.thread_clone.commands.put('!cmd ctemp_disable')

    def read_cryocon(self):
        # !cmd cmeasuretemp     -> reads temperature, saves as CYRO_DEGC
        # !cmd CRYO_DEGC        -> query the CRYO_DEGC variable
        # ?res                  -> gets result of CRYO_DEGC query
        self.thread_clone.commands.put('!cmd cmeasuretemp')

    def mar_enable(self):
        self.thread_clone.commands.put('!cmd mar_enable')
 
    def mar_disable(self):
        self.thread_clone.commands.put('!cmd mar_disable')
 
    def run_loopscan(self,mroot,n_points,exp_time):
        # !cmd loopscan n_points exposure_time sleep_time
        self.thread_clone.commands.put('!cmd mar netroot {}'.format(mroot))
        self.thread_clone.commands.put('!cmd loopscan {} {}'.format(n_points,exp_time))

#------------------------------------------------------------------------------

#    p.addCommand("!cmd slacx_mar_data_path = 'my_mar_data'")
#    p.addCommand("!cmd slacx_pd_filename = 'my_pd_filename'")
#    p.addCommand("!cmd slacx_loopscan_npoints = 2")
#    p.addCommand("!cmd slacx_loopscan_counting_time = 2")
#    p.addCommand("!cmd runme")
#    p.addCommand("?sta")
#
# !rqc                  -> requests control
# !cmd mar_enable       -> enables mar det as a counter
# !cmd pd enable        -> enables pilatus as a counter

# Loop Scanning
# !cmd loopscan n_points exposure_time sleep_time





from collections import OrderedDict

import os.path

from ...Operation import Operation

inputs=OrderedDict(
    file_path=None)
outputs=OrderedDict(
    file_exists_flag=None)

class CheckFilePath(Operation):
    """Checks a file path and returns whether or not the file exists"""

    def __init__(self):
        super(CheckFilePath, self).__init__(inputs, outputs)
        self.input_doc['file_path'] = 'filesystem path'
        self.output_doc['file_exists_flag'] = 'boolean indicating '\
        'whether or not file exists' 

    def run(self):
        fp = self.inputs['file_path']
        self.outputs['file_path'] = os.path.exists(fp)




from collections import OrderedDict

import numpy as np

from ...Operation import Operation

inputs=OrderedDict(
    file_path=None,
    delimiter=None)
outputs=OrderedDict(array=None)

class CSVToArray(Operation):
    """
    Read a csv-formatted file into a numpy array.
    """

    def __init__(self):
        super(CSVToArray, self).__init__(inputs, outputs)
        self.input_doc['file_path'] = "path to .csv file"
        self.output_doc['array'] = "numpy array built from csv file contents"

    def run(self):
        p = self.inputs['file_path']
        self.outputs['array'] = np.loadtxt(p, dtype=float, delimiter=',')





import numpy as np
from collections import OrderedDict

from ...Operation import Operation

inputs=OrderedDict(data_1=None,data_2=None)
outputs=OrderedDict(data_diff=None)

class Subtract(Operation):
    """Subtract one piece of data from another."""

    def __init__(self):
        super(Subtract,self).__init__(inputs,outputs)        
        self.input_doc['data_1'] = 'first piece of data, used as subtrahend'
        self.input_doc['data_2'] = 'second piece of data, used as minuend'
        self.output_doc['data_diff'] = 'the result of subtracting '\
            'data_1 from data_2, i.e. data_diff = data_2 - data_1'

    def run(self):
        self.outputs['data_diff'] = self.inputs['data_2'] - self.inputs['data_1']




from collections import OrderedDict

import numpy as np

import yaml
from ...Operation import Operation

inputs=OrderedDict(file_path=None)
outputs=OrderedDict(yaml_output=None)

class LoadYAML(Operation):
    """Load data from a YAML file

    Returns the output of yaml.load() 
    """

    def __init__(self):
        super(LoadYAML, self).__init__(inputs, outputs)
        self.input_doc['file_path'] = 'path to YAML-formatted data file'
        self.output_doc['yaml_output'] = 'the result of yaml.load(file_path)' 

    def run(self):
        p = self.inputs['file_path']
        f = open(p,'r')
        ds = yaml.load(f)
        f.close()

        #
        #
        #
        #
        for k,v in ds.items():
            if isinstance(v,list) or isinstance(v,np.ndarray):
                ds[k] = float(v[0])
            elif ds[k] is None:
                ds.pop(k)
            else:
                ds[k] = float(v)
        #
        #
        #
        #

        self.outputs['yaml_output'] = ds 






from collections import OrderedDict

import numpy as np

from ...Operation import Operation

inputs = OrderedDict(
    batch_outputs=None,
    x_key=None,
    y_key=None,
    x_sort_flag=False,
    x_shift_flag=False) 
outputs = OrderedDict(
    x=None,
    y=None,
    x_y=None) 

class XYDataFromBatch(Operation):
    """
    Harvest two arrays from a batch output (a list of dicts). 
    Takes a batch output, a key for x values, and a key for y values.
    """

    def __init__(self):
        super(XYDataFromBatch,self).__init__(inputs,outputs)        
        self.input_doc['batch_outputs'] = 'list of dicts produced by a batch execution'
        self.input_doc['x_key'] = 'key for x data from batch_outputs'
        self.input_doc['y_key'] = 'key for y data from batch_outputs'
        self.input_doc['x_sort_flag'] = 'if True, sort data for increasing x' 
        self.input_doc['x_shift_flag'] = 'if True, shift x data so that its minimum value is zero' 
        self.output_doc['x'] = 'array of the x values'
        self.output_doc['y'] = 'array of the y values'
        self.output_doc['x_y'] = 'n-by-2 array of x and y values'

    def run(self):
        b_out = self.inputs['batch_outputs']
        kx = self.inputs['x_key']
        ky = self.inputs['y_key']
        sortflag = self.inputs['x_shift_flag']
        shiftflag = self.inputs['x_shift_flag']

        x_list = []
        y_list = []
        for d in b_out:
            if kx in d and ky in d:
                x_list.append(d[kx])
                y_list.append(d[ky])
        if shiftflag or sortflag:
            x_all = np.array(x_list)
            y_all = np.array(y_list)
        if shiftflag and len(x_list)>0:
            xmin = min(x_list)
            x_all = x_all - xmin
            x_list = list(x_all)
        if sortflag:
            i_xsort = np.argsort(x_all)
            x_list = list(x_all[i_xsort])
            y_list = list(y_all[i_xsort])

        self.outputs['x'] = x_list
        self.outputs['y'] = y_list
        self.outputs['x_y'] = zip(x_list,y_list)





from collections import OrderedDict

import pyFAI

from ...Operation import Operation

inputs=OrderedDict(
    dark_img=None,
    flat_img=None,
    mask_img=None,
    polz_factor=0.)
outputs=OrderedDict(pyfai_args=None)

class GetPyFAIArgs(Operation):
    """Set up a dict of keyword args for a PyFAI.AzimuthalIntegrator."""
    
    def __init__(self):
        super(GetPyFAIArgs,self).__init__(inputs,outputs)
        self.input_doc['dark_img'] = 'image data for dark field'
        self.input_doc['flat_img'] = 'image data for dark field'
        self.input_doc['mask_img'] = 'image data for mask'
        self.input_doc['polz_factor'] = 'polarization factor'
        self.output_doc['poni_dict'] = 'Dict of pyFAI calibration parameters'

    def run(self):
        pdict = dict(
            mask = self.inputs['mask_img'],
            dark = self.inputs['dark_img'],
            flat = self.inputs['flat_img'],
            polarization_factor = self.inputs['polz_factor'])
        self.outputs['pyfai_args'] = pdict 

